/*  
 *  Constructive Solid Geometry program based on information from the
 *  Advanced OpenGL Rendering course at SIGGRAPH '96 by David Blythe and
 *  Tom McReynolds.  Click right mouse button for menu.
 *
 *  author: Nate Robins
 *  email: ndr@pobox.com
 *  www: http://www.pobox.com/~ndr
 */


// includes
#include "glut.h"
#include <math.h>
#include "trackball.h"


// defines
#define SPHERE 1
#define CONE   2
#define CUBE   3


// opration CSG (Constructive Solid Geometry) Traduction: Gomtrie solide constructive
// types d'oprations CSG 
typedef enum {
    CSG_A,
    CSG_B,
    CSG_A_OR_B,
    CSG_A_AND_B,
    CSG_A_SUB_B,
    CSG_B_SUB_A
} csgOperation;


// globals
GLint Width;
GLint Height;

GLfloat zoom = 0.0;

GLboolean cube_picked   = GL_FALSE;
GLboolean sphere_picked = GL_FALSE;
GLboolean cone_picked   = GL_FALSE;

GLboolean selection     = GL_FALSE;		// rendu dans la slection du tampon

#define SELECT_BUFFER 32
GLuint select_buffer[SELECT_BUFFER];	// tampon de slection

GLfloat cone_x = 0.0;
GLfloat cone_y = 0.0;
GLfloat cone_z = 0.0;

GLfloat cube_x = 0.0;
GLfloat cube_y = 0.0;
GLfloat cube_z = 0.0;

GLfloat sphere_x = 0.0;
GLfloat sphere_y = 0.0;
GLfloat sphere_z = 0.0;

GLint mouse_state = -1;
GLint mouse_button = -1;

csgOperation Op = CSG_A_OR_B;

void (*A)(void);
void (*B)(void);


// les fonctions

// un()
// dessiner un seul objet
void one(void(*a)(void))
{
	glEnable(GL_DEPTH_TEST);
	a();
	glDisable(GL_DEPTH_TEST);
}

// ou()
// boolen A ou B (dessiner o A ou B)
// algorithme: simple, il suffit de dessiner les deux avec le test de profondeur activ
void or(void(*a)(void), void(*b)())
{
    glPushAttrib(GL_ALL_ATTRIB_BITS);  //  FAIRE - devrait juste pousser la profondeur
    glEnable(GL_DEPTH_TEST);
    a(); b();
    glPopAttrib();
}

// intrieur()
// dfinit le tampon de gabarit pour montrer la partie de A
// (face avant ou arrire selon 'face')
// qui est  l'intrieur de B
void inside(void(*a)(void), void(*b)(void), GLenum face, GLenum test)
{
    // dessine A dans le tampon de profondeur, mais pas dans le tampon de couleur
    glEnable(GL_DEPTH_TEST);
    glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
    glCullFace(face);
    a();
    
    // utilisez le tampon de pochoir pour trouver les parties de A qui sont  l'intrieur de B
    // en commenant par incrmenter le tampon pochoir partout o les faces avant de B sont...
    glDepthMask(GL_FALSE);
    glEnable(GL_STENCIL_TEST);
    glStencilFunc(GL_ALWAYS, 0, 0);
    glStencilOp(GL_KEEP, GL_KEEP, GL_INCR);
    glCullFace(GL_BACK);
    b();

    // ...puis dcrmenter le tampon de pochoir partout o les faces arrire de B sont
    glStencilOp(GL_KEEP, GL_KEEP, GL_DECR);
    glCullFace(GL_FRONT);
    b();

    // dessine maintenant la partie de A qui est  l'intrieur de B
    glDepthMask(GL_TRUE);
    glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
    glStencilFunc(test, 0, 1);
    glDisable(GL_DEPTH_TEST);
    glCullFace(face);
    a();

    // rinitialiser le test du pochoir
    glDisable(GL_STENCIL_TEST);
}

// correction()
// corrige le tampon de profondeur avec les valeurs de profondeur de A
void fixup(void(*a)(void))
{
    // corrige le tampon de profondeur
    glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
    glEnable(GL_DEPTH_TEST);
    glDisable(GL_STENCIL_TEST);
    glDepthFunc(GL_ALWAYS);
    a();

    // rinitialiser la fonction de profondeur
    glDepthFunc(GL_LESS);
}

// et()
// Boolen A et B (dessinez chaque fois que A coupe B)
// algorithme: trouver o A est  l'intrieur de B, puis trouver o
// B est  l'intrieur de A
void and(void(*a)(void), void(*b)(void))
{
    inside(a, b, GL_BACK, GL_NOTEQUAL);
#if 1 
    fixup(b);
#endif
    inside(b, a, GL_BACK, GL_NOTEQUAL);
}

// soustraction()
// boolen A soustraire B (dessiner partout o A est et B n'est pas)
// algorithme: trouver o A est  l'intrieur de B, puis trouver o
// les faces arrires de B ne sont pas en A
void sub(void(*a)(void), void(*b)(void))
{
    inside(a, b, GL_FRONT, GL_NOTEQUAL);
#if 1
    fixup(b);
#endif
    inside(b, a, GL_BACK, GL_EQUAL);
}

// sphre()
// dessiner une sphre jaune
void sphere(void)
{
    glLoadName(2);
    glPushMatrix();
    glTranslatef(sphere_x, sphere_y, sphere_z);
    glColor3f(1.0, 1.0, 0.0);
    glutSolidSphere(5.0, 16, 16);
    glPopMatrix();
}

// cube()
// dessiner un cube rouge
void cube(void)
{
    glLoadName(1);
    glPushMatrix();
    glTranslatef(cube_x, cube_y, cube_z);
    glColor3f(1.0, 0.0, 0.0);
    glutSolidCube(8.0);
    glPopMatrix();
}

// cne()
// dessiner un cne vert
void cone(void)
{
    glLoadName(3);
    glPushMatrix();
    glTranslatef(cone_x, cone_y, cone_z);
    glColor3f(0.0, 1.0, 0.0);
    glTranslatef(0.0, 0.0, -6.5);
    glutSolidCone(4.0, 15.0, 16, 16);
    glRotatef(180.0, 1.0, 0.0, 0.0);
    glutSolidCone(4.0, 0.0, 16, 1);
    glPopMatrix();
}

void init(void)
{
    GLfloat lightposition[] = { -3.0, 3.0, 3.0, 0.0 };

    tbInit(GLUT_MIDDLE_BUTTON);

    glSelectBuffer(SELECT_BUFFER, select_buffer);

    glDepthFunc(GL_LESS);
    glEnable(GL_DEPTH_TEST);

    glEnable(GL_LIGHT0);
    glEnable(GL_LIGHTING);
    glLightfv(GL_LIGHT0, GL_POSITION, lightposition);
    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);

    glEnable(GL_COLOR_MATERIAL);

    glEnable(GL_CULL_FACE);

    glClearColor(0.0, 0.0, 1.0, 0.0);
}

void reshape(int width, int height)
{
    Width = width;
    Height = height;

    tbReshape(width, height);

    glViewport(0, 0, width, height);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glFrustum(-3.0, 3.0, -3.0, 3.0, 64, 256);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glTranslatef(0.0, 0.0, -200.0 + zoom);
}

void display(void)
{
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

  glPushMatrix();
  tbMatrix();
  
  switch(Op) {
  case CSG_A:
    one(A);
    break;
  case CSG_B:
    one(B);
    break;
  case CSG_A_OR_B:
    or(A, B);
    break;
  case CSG_A_AND_B:
    and(A, B);
    break;
  case CSG_A_SUB_B:
    sub(A, B);
    break;
  case CSG_B_SUB_A:
    sub(B, A);
    break;
  }
  
  glPopMatrix();
  
  if (!selection)
    glutSwapBuffers();
}

GLuint
pick(int x, int y)
{
  GLuint    i, hits, num_names, picked;
  GLuint*   p;
  GLboolean save;
  GLuint    depth = (GLuint)-1;
  GLint     viewport[4];

  //obtenir les paramtres actuels de la fentre
  glGetIntegerv(GL_VIEWPORT, viewport);

  //dfinir le mode de rendu sur la selection
  glRenderMode(GL_SELECT);
  selection = GL_TRUE;
  glInitNames();
  glPushName(0);

  //configurer une matrice de slection et la rendre dans le tampon de slection
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();

  glLoadIdentity();
  gluPickMatrix(x, viewport[3] - y, 5.0, 5.0, viewport);
  glFrustum(-3.0, 3.0, -3.0, 3.0, 64, 256);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glTranslatef(0.0, 0.0, -200.0 + zoom);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

  glPushMatrix();
  tbMatrix();
  or(A, B);
  glPopMatrix();

  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  glMatrixMode(GL_MODELVIEW);

  hits = glRenderMode(GL_RENDER);

  selection = GL_FALSE;

  p = select_buffer;
  picked = 0;
  for (i = 0; i < hits; i++) {
    save = GL_FALSE;
    num_names = *p;			//nombre de noms dans cette frapp
    p++;

    if (*p <= depth) {		//vrifier la 1re valeur de profondeur
      depth = *p;
      save = GL_TRUE;
    }
    p++;
    if (*p <= depth) {		//vrifier la 2me valeur de profondeur
      depth = *p;
      save = GL_TRUE;
    }
    p++;

    if (save)
      picked = *p;

    p += num_names;			//passer le reste des noms
  }

  return picked;
}

void keyboard(unsigned char key, int x, int y)
{
    switch(key) {
    case 'c':
	if(A == cube && B == sphere) {
	    A = sphere;
	    B = cone;
	} else if(A == sphere && B == cone) {
	    A = cone;
	    B = cube;
	} else {				//si (A == cne && B = cube)
	    A = cube;
	    B = sphere;
	}
	break;
    case 'a':
	Op = CSG_A;
	break;
    case 'b':
	Op = CSG_B;
	break;
    case 'o':
	Op = CSG_A_OR_B;
	break;
    case 'e':
	Op = CSG_A_AND_B;
	break;
    case '-':
	Op = CSG_A_SUB_B;
	break;
    case '_':
	Op = CSG_B_SUB_A;
	break;
    case 'z':
	zoom -= 6.0;
	reshape(Width, Height);
	break;
    case 'Z':
	zoom += 6.0;
	reshape(Width, Height);
	break;
    case 27:
	exit(0);
	break;
    case '\r':
	break;
    default:
	return;
    }

    glutPostRedisplay();
}

void
mouse(int button, int state, int x, int y)
{
  mouse_state = state;
  mouse_button = button;

  tbMouse(button, state, x, y);

  if (button == GLUT_LEFT_BUTTON) {
    switch (pick(x, y)) {
    case 1:
      cube_picked = GL_TRUE;
      sphere_picked = GL_FALSE;
      cone_picked = GL_FALSE;
      break;
      
    case 2:
      sphere_picked = GL_TRUE;
      cube_picked = GL_FALSE;
      cone_picked = GL_FALSE;
      break;
      
    case 3:
      cone_picked = GL_TRUE;
      sphere_picked = GL_FALSE;
      cube_picked = GL_FALSE;
      break;
      
    default:
      sphere_picked = GL_FALSE;
      cube_picked = GL_FALSE;
      cone_picked = GL_FALSE;
      break;
    }
  }

  glutPostRedisplay();
}

GLvoid
cross(GLfloat* u, GLfloat* v, GLfloat* n)
{
  //calculer le produit crois (u x v pour droitier [ccw])
  n[0] = u[1] * v[2] - u[2] * v[1];
  n[1] = u[2] * v[0] - u[0] * v[2];
  n[2] = u[0] * v[1] - u[1] * v[0];
}

GLfloat
normalize(GLfloat* n)
{
  GLfloat l;

  //normaliser
  l = (GLfloat)sqrt(n[0] * n[0] + n[1] * n[1] + n[2] * n[2]);
  n[0] /= l;
  n[1] /= l;
  n[2] /= l;

  return l;
}

void
motion(int x, int y)
{
  GLdouble model[4*4];
  GLdouble proj[4*4];
  GLint view[4];
  GLdouble pan_x, pan_y, pan_z;

  tbMotion(x, y);

  if (mouse_state == GLUT_DOWN && mouse_button == GLUT_LEFT_BUTTON) {
    glGetDoublev(GL_MODELVIEW_MATRIX, model);
    glGetDoublev(GL_PROJECTION_MATRIX, proj);
    glGetIntegerv(GL_VIEWPORT, view);
    gluProject((GLdouble)x, (GLdouble)y, 0.0,
		 model, proj, view,
		 &pan_x, &pan_y, &pan_z);
    gluUnProject((GLdouble)x, (GLdouble)y, pan_z,
		 model, proj, view,
		 &pan_x, &pan_y, &pan_z);
    pan_y = -pan_y;

    if (sphere_picked) {
      sphere_x = pan_x;
      sphere_y = pan_y;
      sphere_z = pan_z;
    } else if (cone_picked) {
      cone_x = pan_x;
      cone_y = pan_y;
      cone_z = pan_z;
    } else if (cube_picked) {
      cube_x = pan_x;
      cube_y = pan_y;
      cube_z = pan_z;
    }
  }

  glutPostRedisplay();
}

void menu(int item)
{
    keyboard((unsigned char)item, 0, 0);
}

void main(int argc, char** argv)
{
    int ops, zoom;

    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH | GLUT_STENCIL);

    glutCreateWindow("Dmo Oprations CSG");

    glutDisplayFunc(display);
    glutReshapeFunc(reshape);
    glutKeyboardFunc(keyboard);
    glutMouseFunc(mouse);
    glutMotionFunc(motion);

    ops = glutCreateMenu(menu);
    glutAddMenuEntry("A Seulement          (a)", 'a');
    glutAddMenuEntry("B seulement          (b)", 'b');
    glutAddMenuEntry("A ou B                    (o)", 'o');
    glutAddMenuEntry("A et B                      (e)", 'e');
    glutAddMenuEntry("A moin B                (-)", '-');
    glutAddMenuEntry("B moin A                (_)", '_');
    zoom = glutCreateMenu(menu);
    glutAddMenuEntry("Diminution du zoom       (z)", 'z');
    glutAddMenuEntry("Augmentation du zoom (Z)", 'Z');
    glutCreateMenu(menu);
    glutAddMenuEntry("Dmo Oprations CSG     ", '\0');
    glutAddMenuEntry("                        ", '\0');
    glutAddSubMenu(  "Oprations              ", ops);
    glutAddSubMenu(  "Zoom                    ", zoom);
    glutAddMenuEntry("                        ", '\0');
    glutAddMenuEntry("Changer forme   (c)     ", 'c');
    glutAddMenuEntry("                        ", '\0');
    glutAddMenuEntry("Quitter     (chap)     ", '\033');
    glutAttachMenu(GLUT_RIGHT_BUTTON);

    init();

    A = cube;
    B = sphere;

    glutMainLoop();
}
